package hardcaml_xilinx_reports

  1. Overview
  2. Docs

Module Hardcaml_xilinx_reports.ProjectSource

Vivado synthesis project generation.

Sourcemodule Config : sig ... end

Project configuration.

Sourcetype t
Sourceval create : ?database:Hardcaml.Circuit_database.t -> ?config:Config.t -> ?place:Base.Bool.t -> ?route:Base.Bool.t -> ?checkpoint:Base.Bool.t -> clocks:Clock.t Base.List.t -> part_name:Base.String.t -> output_path:Base.String.t -> Hardcaml.Circuit.t -> t

Create a Vivado project. This consists of the RTL generated for the given Circuit.t and project files. The function should be passed a list of top level clocks which specify the port name and requested frequency. part_name is the full FPGA part name including package a speed-grade. Files will be written to output_path. Performs synthesis by default but will also optionally run placement and routing.

Sourceval run : ?verbose:Base.Bool.t -> ?path_to_vivado:Base.String.t -> t -> Report.t Base.Option.t Async.Deferred.t

Execute a project generated with create.

Uses Unix.system to run vivado in batch mode.

Sourceval output_path : t -> Base.String.t

Output path where reports, artifacts and verilog files for this project lives.